# Import opinel
from opinel.utils import *

# Import Scout2 tools
from AWSScout2.utils import *


########################################
##### Globals
########################################

fetched_alarms = 0
discovered_alarms = 0
fetched_metrics = 0
discovered_metrics = 0


########################################
##### CloudWatch functions
########################################

#
# Get CloudWatch config in all regions in scope
#
def get_cloudwatch_info(credentials, service_config, selected_regions, partition_name):
    manage_dictionary(service_config, 'regions', {})
    printInfo('Fetching CloudWatch config...')
    cloudwatch_status_init()
    for region in build_region_list('cloudwatch', selected_regions, partition_name):
        manage_dictionary(service_config['regions'], region, {})
        service_config['regions'][region]['name'] = region
    thread_work(service_config['regions'], threaded_per_region, params = {'method': get_cloudwatch_region, 'creds': credentials, 'cloudwatch_config': service_config})
    service_config['regions_count'] = len(service_config['regions'])
    cloudwatch_status(True)


#
# Get CloudWatch config in a single region
#
def get_cloudwatch_region(params):
    global fetched_alarms, discovered_alarms, fetched_metrics, discovered_metrics
    region = params['region']
    cloudwatch_config = params['cloudwatch_config']
    manage_dictionary(cloudwatch_config['regions'][region], 'metrics', {})
    cloudwatch_client = connect_service('cloudwatch', params['creds'], region)
    metrics = {}
    cloudwatch_config['regions'][region]['alarms'] = {}
    alarms = handle_truncated_response(cloudwatch_client.describe_alarms, {}, ['MetricAlarms'])['MetricAlarms']
    cloudwatch_config['regions'][region]['alarms_count'] = len(alarms)
    discovered_alarms += len(alarms)
    cloudwatch_status()
    for alarm in alarms:
        alarm['arn'] = alarm.pop('AlarmArn')
        alarm['name'] = alarm.pop('AlarmName')
        # Drop some data 
        for k in ['AlarmConfigurationUpdatedTimestamp', 'StateReason', 'StateReasonData', 'Statistic', 'ComparisonOperator', 'Threshold', 'StateUpdatedTimestamp']:
            foo = alarm.pop(k) if k in alarm else None
        alarm_id = get_non_aws_id(alarm['arn'])
        metric_id = get_non_aws_id('%s/%s' % (alarm['Namespace'], alarm['MetricName']))
        cloudwatch_config['regions'][region]['alarms'][alarm_id] = alarm
        manage_dictionary(metrics, metric_id, {'name': alarm['MetricName'],        'namespace': alarm['Namespace']})
        fetched_alarms += 1
        cloudwatch_status()
    cloudwatch_config['regions'][region]['alarms'] = alarms
    cloudwatch_config['regions'][region]['metrics'] = metrics
    cloudwatch_config['regions'][region]['metrics_count'] = len(metrics)
    #printInfo('Fetching metrics in %s...' % region)
    #metrics = handle_truncated_response(cloudwatch_client.list_metrics, {}, ['Metrics'])['Metrics']
    #for metric in metrics:
    #cloudwatch_config['regions'][region]['metrics_count'] = len(metrics)
    #cloudwatch_config['regions'][region]['metrics'] = metrics


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

try:
    from opinel.utils import check_opinel_version, configPrintException, get_opinel_requirement, printInfo, read_creds
except Exception as e:
    print('Error: Scout2 depends on the opinel package. Install all the requirements with the following command:')
    print('  $ pip install -r requirements.txt')
    print(e)
    sys.exit(42)

from AWSScout2 import AWSCONFIG, __version__ as scout2_version
from AWSScout2.cli_parser import Scout2ArgumentParser
from AWSScout2.configs.scout2 import  Scout2Config
from AWSScout2.configs.services import postprocessing
from AWSScout2.output.html import Scout2Report
from AWSScout2.rules.exceptions import process_exceptions
from AWSScout2.rules.ruleset import Ruleset
from AWSScout2.rules.postprocessing import do_postprocessing


########################################
##### Main
########################################

def main():

    # Parse arguments
    parser = Scout2ArgumentParser()
    args = parser.parse_args()

    # Configure the debug level
    configPrintException(args.debug)

    # Check version of opinel
    min_opinel, max_opinel = get_opinel_requirement(os.path.realpath(__file__))
    if not check_opinel_version(min_opinel):
        return 42

    # Set the profile name
    profile_name = args.profile[0]

    # Search for AWS credentials
    if not args.fetch_local:
        credentials = read_creds(args.profile[0], args.csv_credentials, args.mfa_serial, args.mfa_code)
        if credentials['AccessKeyId'] is None:
            return 42

    # Create a new Scout2 config
    report = Scout2Report(profile_name, args.report_dir, args.timestamp)
    aws_config = Scout2Config(profile_name, args.report_dir, args.timestamp, args.services, args.skipped_services)

    # Fetch data from AWS APIs if not running a local analysis
    if not args.fetch_local:

        aws_config.fetch(credentials, regions=args.regions, partition_name=args.partition_name)
        aws_config.update_metadata()
        report.save(aws_config, {}, None, args.force_write, args.debug)

    # Reload to flatten everything into a python dictionary
    aws_config = report.jsrw.load_from_file(AWSCONFIG)

    # Analyze config
    ruleset = Ruleset(profile_name)
    ruleset.analyze(aws_config)

    # Create display filters
    filters = Ruleset(filename = 'filters.json', rule_type = 'filters')
    filters.analyze(aws_config)

    # Finalize
    postprocessing(aws_config)
    do_postprocessing(aws_config)


    # Handle exceptions
    process_exceptions(aws_config, args.exceptions[0])

    # Save config and create HTML report
    last_run = {}
    last_run['time'] = report.current_time.strftime("%Y-%m-%d %H:%M:%S%z")
    last_run['cmd'] = ' '.join(sys.argv)
    last_run['version'] = scout2_version
    last_run['ruleset_name'] = ruleset.name
    last_run['ruleset_about'] = ruleset.ruleset['about'] if 'about' in ruleset.ruleset else ''
    report.save(aws_config, {}, last_run, args.force_write, args.debug)


# -*- coding: utf-8 -*-
"""
Redshift-related classes and functions
"""

# Import opinel
from opinel.utils import handle_truncated_response, manage_dictionary

# Import AWSScout2
from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig, api_clients



########################################
# RedshiftRegionConfig
########################################

class RedshiftRegionConfig(RegionConfig):
    """
    Redshift configuration for a single AWS region

    :ivar vpcs:                         Dictionary of VPCs [id]
    :ivar clusters_count:               Number of clusters in the region
    :ivar parameter_groups:             Dictionary of parameter groups [id]
    :ivar parameter_groups_count:       Number of parameter groups in the region
    :ivar security_groups:              Dictionary of security groups [id]
    :ivar security_groups_count:        Number of security groups in the region
    """

    def __init__(self):
        self.vpcs = {}
        self.clusters_count = 0
        self.parameter_groups = {}
        self.parameter_groups_count = 0
        self.security_groups = {}
        self.security_groups_count = 0


    def parse_cluster(self, global_params, region, cluster):
        """
        Parse a single Redshift cluster

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param cluster:                 Cluster
        """
        vpc_id = cluster.pop('VpcId') if 'VpcId' in cluster else ec2_classic
        manage_dictionary(self.vpcs, vpc_id, RedshiftVPCConfig())
        name = cluster.pop('ClusterIdentifier')
        cluster['name'] = name
        self.vpcs[vpc_id].clusters[name] = cluster


    def parse_parameter_group(self, global_params, region, parameter_group):
        """
        Parse a single Redshift parameter group and fetch all of its parameters

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param parameter_group:         Parameter group
        """
        pg_name = parameter_group.pop('ParameterGroupName')
        pg_id = self.get_non_aws_id(pg_name) # Name could be used as only letters digits or hyphens
        parameter_group['name'] = pg_name
        parameter_group['parameters'] = {}
        api_client = api_clients[region]
        parameters = handle_truncated_response(api_client.describe_cluster_parameters, {'ParameterGroupName': pg_name}, ['Parameters'])['Parameters']
        for parameter in parameters:
            param = {}
            param['value'] = parameter['ParameterValue']
            param['source'] = parameter['Source']
            parameter_group['parameters'][parameter['ParameterName']] = param
        (self).parameter_groups[pg_id] = parameter_group


    def parse_security_group(self, global_params, region, security_group):
        """
        Parse a single Redsfhit security group

        :param global_params:           Parameters shared for all regions
        :param region:                  Name of the AWS region
        :param security)_group:         Security group
        """
        name = security_group.pop('ClusterSecurityGroupName')
        security_group['name'] = name
        self.security_groups['name'] = security_group



########################################
# RedshiftConfig
########################################

class RedshiftConfig(RegionalServiceConfig):
    """
    Redshift configuration for all AWS regions

    :cvar targets:                      Tuple with all Redshift resource names that may be fetched
    :cvar config_class:                 Class to be used when initiating the service's configuration in a new region/VPC
    """
    targets = (
        ('clusters', 'Clusters', 'describe_clusters', False),
        ('parameter_groups', 'ParameterGroups', 'describe_cluster_parameter_groups', False),
        ('security_groups', 'SecurityGroups', 'describe_cluster_security_groups', True),
        # TODO ('subnets')
    )
    region_config_class = RedshiftRegionConfig



########################################
# RedshiftVPCConfig
########################################

class RedshiftVPCConfig(object):
    """
    Redshift configuration for a single VPC

    :ivar clusters:                     Dictionary of clusters [name]
    """

    def __init__(self):
        self.clusters = {}


# -*- coding: utf-8 -*-

from opinel.utils import manage_dictionary

from opinel.utils import load_data
from opinel.utils_ec2 import get_name

#from AWSScout2.utils import *
from AWSScout2.utils import ec2_classic, get_keys
from AWSScout2.configs.regions import RegionalServiceConfig, RegionConfig, api_clients



########################################
# Globals
########################################

protocols_dict = load_data('protocols.json', 'protocols')



########################################
# VPCRegionConfig
########################################

class VPCRegionConfig(RegionConfig):
    """
    VPC configuration for a single AWS region

    :ivar subnets:                       Dictionary of subnets [name]
    :ivar subnets_count:                 Number of subnets in the region
    """

    def __init__(self):
        self.flow_logs_count = 0
        self.network_acls_count = 0
        self.subnets = {}               # Temporary artifact
        self.subnets_count = 0
        self.vpcs = {}
        # Gateways etc... (vpn, internet, nat, ...)


    def parse_flow_log(self, global_params, region, fl):
        """

        :param global_params:
        :param region:
        :param fl:
        :return:
        """
        fl_id = fl.pop('FlowLogId')
        resource_id = fl.pop('ResourceId')
        if resource_id.startswith('vpc-'):
            # TODO TODO TODO region_info no longer works
            manage_dictionary(region_info['vpcs'], resource_id, {})
            manage_dictionary(region_info['vpcs'][resource_id], 'flow_logs', {})
            manage_dictionary(self.vpcs, resource_id, SingleVPCConfig())
            self.vpcs[resource_id]['flow_logs'][fl_id] = fl
        elif resource_id.startswith('subnet-'):
            # Temporary save within the region, once subnets are fetched too, do an update at the end of the run
            manage_dictionary(region_info, 'subnets', {})
            manage_dictionary(region_info['subnets'], resource_id, {})
            manage_dictionary(region_info['subnets'][resource_id], 'flow_logs', {})
            self.subnets[resource_id]['flow_logs'][fl_id] = fl


    def parse_network_acl(self, global_params, region, network_acl):
        """

        :param global_params:
        :param region:
        :param network_acl:
        :return:
        """
        vpc_id = network_acl['VpcId']
        network_acl['id'] = network_acl.pop('NetworkAclId')
        get_name(network_acl, network_acl, 'id')
        manage_dictionary(network_acl, 'rules', {})
        network_acl['rules']['ingress'] = self.__parse_network_acl_entries(network_acl['Entries'], False)
        network_acl['rules']['egress'] = self.__parse_network_acl_entries(network_acl['Entries'], True)
        network_acl.pop('Entries')
        # Save
        manage_dictionary(self.vpcs, vpc_id, SingleVPCConfig())
        self.vpcs[vpc_id].network_acls[network_acl['id']] = network_acl


    def __parse_network_acl_entries(self, entries, egress):
        """

        :param entries:
        :param egress:
        :return:
        """
        acl_list = []
        for entry in entries:
            if entry['Egress'] == egress:
                acl = {}
                for key in ['CidrBlock', 'RuleAction', 'RuleNumber']:
                    acl[key] = entry[key]
                acl['protocol'] = protocols_dict[entry['Protocol']]
                if 'PortRange' in entry:
                    from_port = entry['PortRange']['From'] if entry['PortRange']['From'] else 1
                    to_port = entry['PortRange']['To'] if entry['PortRange']['To'] else 65535
                    acl['port_range'] = from_port if from_port == to_port else str(from_port) + '-' + str(to_port)
                else:
                    acl['port_range'] = '1-65535'

                acl_list.append(acl)
        return acl_list


    def parse_route_table(self, global_params, region, rt):
        route_table = {}
        vpc_id = rt['VpcId']
        get_name(rt, route_table, 'VpcId') # TODO: change get_name to have src then dst
        get_keys(rt, route_table, ['Routes', 'Associations', 'PropagatingVgws'])
        # Save
        manage_dictionary(self.vpcs, vpc_id, SingleVPCConfig())
        self.vpcs[vpc_id].route_tables[rt['RouteTableId']] = route_table


    def parse_subnet(self, global_params, region, subnet):
        """

        :param global_params:
        :param region:
        :param subnet:
        :return:
        """
        vpc_id = subnet['VpcId']
        manage_dictionary(self.vpcs, vpc_id, SingleVPCConfig())
        subnet_id = subnet['SubnetId']
        get_name(subnet, subnet, 'SubnetId')
        # Save
        manage_dictionary(self.vpcs, vpc_id, SingleVPCConfig())
        self.vpcs[vpc_id].subnets[subnet_id] = subnet


    def parse_vpc(self, global_params, region_name, vpc):
        """

        :param global_params:
        :param region_name:
        :param vpc:
        :return:
        """
        vpc_id = vpc['VpcId']
        # Save
        manage_dictionary(self.vpcs, vpc_id, SingleVPCConfig())
        self.vpcs[vpc_id].name = get_name(vpc, {}, 'VpcId')



########################################
# VPCConfig
########################################

class VPCConfig(RegionalServiceConfig):
    """
    VPC configuration for all AWS regions

    :cvar targets:                      Tuple with all VPC resource names that may be fetched
    :cvar region_config_class:          Class to be used when initiating the service's configuration in a new region
    """
    targets = (
        ('vpcs', 'Vpcs', 'describe_vpcs', False),
        ('flow_logs', 'FlowLogs', 'describe_flow_logs', False),
        ('network_acls', 'NetworkAcls', 'describe_network_acls', False),
        ('route_tables', 'RouteTables', 'describe_route_tables', False),
        ('subnets', 'Subnets', 'describe_subnets', False)
    )
    region_config_class = VPCRegionConfig



########################################
# SingleVPCConfig
########################################

class SingleVPCConfig(object):
    """
    Configuration for a single VPC

    :ivar flow_logs:                    Dictionary of flow logs [id]
    :ivar instances:                    Dictionary of instances [id]
    """

    def __init__(self, name = None):
        self.name = name
        self.flow_logs = {}
        self.network_acls = {}
        self.route_tables = {}
        self.subnets = {}




















########################################
##### VPC analysis functions
########################################



#
# Add a display name for all known CIDRs
#
known_cidrs = {'0.0.0.0/0': 'All'}
def put_cidr_name(aws_config, current_config, path, current_path, resource_id, callback_args):
    if 'cidrs' in current_config:
        cidr_list = []
        for cidr in current_config['cidrs']:
            if type(cidr) == dict:
                cidr = cidr['CIDR']
            if cidr in known_cidrs:
                cidr_name = known_cidrs[cidr]
            else:
                cidr_name = get_cidr_name(cidr, callback_args['ip_ranges'], callback_args['ip_ranges_name_key'])
                known_cidrs[cidr] = cidr_name
            cidr_list.append({'CIDR': cidr, 'CIDRName': cidr_name})
        current_config['cidrs'] = cidr_list

#
# Read display name for CIDRs from ip-ranges files
#
#aws_ip_ranges = read_ip_ranges(aws_ip_ranges_filename, False)
def get_cidr_name(cidr, ip_ranges_files, ip_ranges_name_key):
    for filename in ip_ranges_files:
        ip_ranges = read_ip_ranges(filename, local_file = True)
        for ip_range in ip_ranges:
            ip_prefix = netaddr.IPNetwork(ip_range['ip_prefix'])
            cidr = netaddr.IPNetwork(cidr)
            if cidr in ip_prefix:
                return ip_range[ip_ranges_name_key]
    for ip_range in aws_ip_ranges:
        ip_prefix = netaddr.IPNetwork(ip_range['ip_prefix'])
        cidr = netaddr.IPNetwork(cidr)
        if cidr in ip_prefix:
            return 'Unknown CIDR in %s %s' % (ip_range['service'], ip_range['region'])
    return 'Unknown CIDR'

#
# Propagate VPC names in VPC-related services (info only fetched during EC2 calls)
#
def propagate_vpc_names(aws_config, current_config, path, current_path, resource_id, callback_args):
    if resource_id == ec2_classic:
        current_config['name'] = ec2_classic
    else:
        target_path = copy.deepcopy(current_path)
        target_path[1] = 'ec2'
        target_path.append(resource_id)
        target_path.append('Name')
        target_path = '.'.join(target_path)
        current_config['name'] = get_value_at(aws_config, target_path, target_path)



# -*- coding: utf-8 -*-
"""
Multi-service post-processing functions
"""

from opinel.utils import manage_dictionary, printInfo, printError, printException
from AWSScout2.utils import ec2_classic

from AWSScout2.configs.browser import get_object_at

import copy

def do_postprocessing(aws_config):
    match_instances_and_roles(aws_config)
    match_iam_policies_and_buckets(aws_config)
    vpc_postprocessing(aws_config)
    # final
    update_metadata(aws_config)







#
# Create dashboard metadata
#
def update_metadata(aws_config):
    # Security risks dropdown on a per-resource basis

    service_map = {}
    for service_group in aws_config['metadata']:
        for service in aws_config['metadata'][service_group]:
            if service not in aws_config['service_list']:
                continue
            if 'resources' not in aws_config['metadata'][service_group][service]:
                continue
            service_map[service] = service_group
    for s in aws_config['services']:
        if aws_config['services'][s] and 'violations' in aws_config['services'][s]:
            for v in aws_config['services'][s]['violations']:
                # Finding resource
                resource_path = aws_config['services'][s]['violations'][v]['display_path'] if 'display_path' in aws_config['services'][s]['violations'][v] else aws_config['services'][s]['violations'][v]['path']
                resource = resource_path.split('.')[-2]
                # h4ck...
                if resource == 'credential_report':
                    resource = resource_path.split('.')[-1].replace('>', '').replace('<', '')
                elif resource == s:
                    resource = resource_path.split('.')[-1]
                if aws_config['services'][s]['violations'][v]['flagged_items'] > 0:
                    try:
                        manage_dictionary(aws_config['metadata'][service_map[s]][s]['resources'][resource], 'risks', [])
                        aws_config['metadata'][service_map[s]][s]['resources'][resource]['risks'].append(v)
                    except Exception as e:
                        try:
                            manage_dictionary(aws_config['metadata'][service_map[s]][s]['summaries'][resource], 'risks', [])
                            aws_config['metadata'][service_map[s]][s]['summaries'][resource]['risks'].append(v)
                        except Exception as e:
                            printError('Service: %s' % s)
                            printError('Resource: %s' % resource)
                            printException(e)


########################################
# EC2 & IAM
########################################

def match_instances_and_roles(aws_config):
    """
    Foobar

    :param aws_config:
    """
    printInfo('Matching EC2 instances and IAM roles...')
    ec2_config = aws_config['services']['ec2']
    iam_config = aws_config['services']['iam']
    role_instances = {}
    for r in ec2_config['regions']:
        for v in ec2_config['regions'][r]['vpcs']:
            if 'instances' in ec2_config['regions'][r]['vpcs'][v]:
                for i in ec2_config['regions'][r]['vpcs'][v]['instances']:
                    instance_profile_id = ec2_config['regions'][r]['vpcs'][v]['instances'][i]['iam_instance_profile']['id'] if 'iam_instance_profile' in ec2_config['regions'][r]['vpcs'][v]['instances'][i] else None
                    if instance_profile_id:
                        manage_dictionary(role_instances, instance_profile_id, [])
                        role_instances[instance_profile_id].append(i)
    for role_id in iam_config['roles']:
        iam_config['roles'][role_id]['instances_count'] = 0
        for instance_profile_id in iam_config['roles'][role_id]['instance_profiles']:
            if instance_profile_id in role_instances:
                iam_config['roles'][role_id]['instance_profiles'][instance_profile_id]['instances'] = role_instances[instance_profile_id]
                iam_config['roles'][role_id]['instances_count'] += len(role_instances[instance_profile_id])



########################################
# S3 & IAM
########################################

def match_iam_policies_and_buckets(aws_config):
    printInfo('Matching S3 buckets and IAM policies...')
    s3_info = aws_config['services']['s3']
    iam_info = aws_config['services']['iam']
    if 'Action' in iam_info['permissions']:
        for action in (x for x in iam_info['permissions']['Action'] if ((x.startswith('s3:') and x != 's3:ListAllMyBuckets') or (x == '*'))):
            for iam_entity in iam_info['permissions']['Action'][action]:
                if 'Allow' in iam_info['permissions']['Action'][action][iam_entity]:
                    for allowed_iam_entity in iam_info['permissions']['Action'][action][iam_entity]['Allow']:
                        # For resource statements, we can easily rely on the existing permissions structure
                        if 'Resource' in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]:
                            for full_path in (x for x in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['Resource'] if x.startswith('arn:aws:s3:') or x == '*'):
                                parts = full_path.split('/')
                                bucket_name = parts[0].split(':')[-1]
                                __update_iam_permissions(s3_info, bucket_name, iam_entity, allowed_iam_entity, iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['Resource'][full_path])
                        # For notresource statements, we must fetch the policy document to determine which buckets are not protected
                        if 'NotResource' in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]:
                            for full_path in (x for x in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'] if x.startswith('arn:aws:s3:') or x == '*'):
                                for policy_type in ['InlinePolicies', 'ManagedPolicies']:
                                    if policy_type in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'][full_path]:
                                        for policy in iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'][full_path][policy_type]:
                                            __update_bucket_permissions(s3_info, iam_info, action, iam_entity, allowed_iam_entity, full_path, policy_type, policy)

def __update_bucket_permissions(s3_info, iam_info, action, iam_entity, allowed_iam_entity, full_path, policy_type, policy_name):
    allowed_buckets = []
    # By default, all buckets are allowed
    for bucket_name in s3_info['buckets']:
        allowed_buckets.append(bucket_name)
    if policy_type == 'InlinePolicies':
        policy = iam_info[iam_entity.title()][allowed_iam_entity]['Policies'][policy_name]['PolicyDocument']
    elif policy_type == 'ManagedPolicies':
        policy = iam_info['ManagedPolicies'][policy_name]['PolicyDocument']
    else:
        printError('Error, found unknown policy type.')
    for statement in policy['Statement']:
        for target_path in statement['NotResource']:
            parts = target_path.split('/')
            bucket_name = parts[0].split(':')[-1]
            path = '/' + '/'.join(parts[1:]) if len(parts) > 1 else '/'
            if (path == '/' or path == '/*') and (bucket_name in allowed_buckets):
                # Remove bucket from list
                allowed_buckets.remove(bucket_name)
            elif bucket_name == '*':
                allowed_buckets = []
    policy_info = {}
    policy_info[policy_type] = {}
    policy_info[policy_type][policy_name] = iam_info['permissions']['Action'][action][iam_entity]['Allow'][allowed_iam_entity]['NotResource'][full_path][policy_type][policy_name]
    for bucket_name in allowed_buckets:
        __update_iam_permissions(s3_info, bucket_name, iam_entity, allowed_iam_entity, policy_info)

def __update_iam_permissions(s3_info, bucket_name, iam_entity, allowed_iam_entity, policy_info):
    if bucket_name != '*' and bucket_name in s3_info['buckets']:
        bucket = s3_info['buckets'][bucket_name]
        manage_dictionary(bucket, iam_entity, {})
        manage_dictionary(bucket, iam_entity + '_count', 0)
        if not allowed_iam_entity in bucket[iam_entity]:
            bucket[iam_entity][allowed_iam_entity] = {}
            bucket[iam_entity + '_count'] = bucket[iam_entity + '_count'] + 1

        if 'inline_policies' in policy_info:
            manage_dictionary(bucket[iam_entity][allowed_iam_entity], 'inline_policies', {})
            bucket[iam_entity][allowed_iam_entity]['inline_policies'].update(policy_info['inline_policies'])
        if 'policies' in policy_info:
            manage_dictionary(bucket[iam_entity][allowed_iam_entity], 'policies', {})
            bucket[iam_entity][allowed_iam_entity]['policies'].update(policy_info['policies'])
    elif bucket_name == '*':
        for bucket in s3_info['buckets']:
            __update_iam_permissions(s3_info, bucket, iam_entity, allowed_iam_entity, policy_info)
        pass
    else:
        # Could be an error or cross-account access, ignore...
        pass



########################################
# VPC
########################################

def vpc_postprocessing(aws_config, ip_ranges = [], ip_ranges_name_key = None):
    printInfo('Post-processing VPC config...')
    # Security group usage: EC2 instances
    callback_args = {'status_path': [], 'sg_list_attribute_name': 'Groups', 'sg_id_attribute_name': 'GroupId'}
    go_to_and_do(aws_config, aws_config['services']['ec2'], ['regions', 'vpcs', 'instances', 'network_interfaces'], ['services', 'ec2'], list_resources_in_security_group, callback_args)
    # Security group usage: ELBs
    callback_args = {'status_path': ['Scheme'], 'sg_list_attribute_name': 'security_groups', 'sg_id_attribute_name': 'GroupId'}
    go_to_and_do(aws_config, aws_config['services']['ec2'], ['regions', 'vpcs', 'elbs'], ['services', 'ec2'], list_resources_in_security_group, callback_args)
    # Security group usage: Redshift clusters
    callback_args = {'status_path': ['ClusterStatus'], 'sg_list_attribute_name': 'VpcSecurityGroups', 'sg_id_attribute_name': 'VpcSecurityGroupId'}
    go_to_and_do(aws_config, aws_config['services']['redshift'], ['regions', 'vpcs', 'clusters'], ['services', 'redshift'], list_resources_in_security_group, callback_args)
    # Security group usage: RDS instances
    callback_args = {'status_path': ['DBInstanceStatus'], 'sg_list_attribute_name': 'VpcSecurityGroups', 'sg_id_attribute_name': 'VpcSecurityGroupId'}
    go_to_and_do(aws_config, aws_config['services']['rds'], ['regions', 'vpcs', 'instances'], ['services', 'rds'], list_resources_in_security_group, callback_args)
    # Add friendly name for CIDRs
    if len(ip_ranges):
        callback_args = {'ip_ranges': ip_ranges, 'ip_ranges_name_key': ip_ranges_name_key}
        go_to_and_do(aws_config, aws_config['services']['ec2'], ['regions', 'vpcs', 'security_groups', 'rules', 'protocols', 'ports'], ['services', 'ec2'], put_cidr_name, callback_args)
    # Propagate VPC names outside EC2
    vpc_services = [ 'rds', 'redshift' ]
#    for service in vpc_services:
#        go_to_and_do(aws_config, aws_config['services'][service], ['regions', 'vpcs'], ['services', service], propagate_vpc_names, {})
    # Remove empty EC2-classic VPC placeholders
    go_to_and_do(aws_config, aws_config['services']['ec2'], ['regions'], ['services', 'ec2'], remove_empty_ec2_placeholders, {})


#
# Recursively go to a target and execute a callback
#
def go_to_and_do(aws_config, current_config, path, current_path, callback, callback_args = None):
  try:
    key = path.pop(0)
    if not current_config:
        current_config = aws_config
    if not current_path:
        current_path = []
    keys = key.split('.')
    if len(keys) > 1:
        while True:
            key = keys.pop(0)
            if not len(keys):
                break
            current_path.append(key)
            current_config = current_config[key]
    if key in current_config:
        current_path.append(key)
        for (i, value) in enumerate(current_config[key]):
            if len(path) == 0:
                if type(current_config[key] == dict) and type(value) != dict and type(value) != list:
                    callback(aws_config, current_config[key][value], path, current_path, value, callback_args)
                else:
                    # TODO: the current_config value passed here is not correct...
                    callback(aws_config, current_config, path, current_path, value, callback_args)
            else:
                tmp = copy.deepcopy(current_path)
                try:
                    tmp.append(value)
                    go_to_and_do(aws_config, current_config[key][value], copy.deepcopy(path), tmp, callback, callback_args)
                except:
                    tmp.pop()
                    tmp.append(i)
                    go_to_and_do(aws_config, current_config[key][i], copy.deepcopy(path), tmp, callback, callback_args)

  except Exception as e:
    printException(e)
    printInfo('Index: %s' % str(i))
    printInfo('Path: %s' % str(current_path))
#    printInfo('Config: %s' % str(current_config))
    printInfo('Key = %s' % str(key))
    printInfo('Value = %s' % str(value))
    printInfo('Path = %s' % path)


#
# EC2-Classic placeholders were automatically created - remove if empty after fetching all the data
#
def remove_empty_ec2_placeholders(aws_config, current_config, path, current_path, resource_id, callback_args):
    if ec2_classic in current_config and current_config['vpcs'][ec2_classic] == {}:
        current_config['vpcs'].pop(ec2_classic)




#
# List the resources associated with a given VPC security group (e.g. ec2 instances, redshift clusters, ...)
# TODO: fix it when single region is fetched but more region data exist...
#
def list_resources_in_security_group(aws_config, current_config, path, current_path, resource_id, callback_args):
    # Retrieve service and resource type from current path
    service = current_path[1]
    resource_type = current_path[-1]
    # Get resource
    resource_path = copy.deepcopy(current_path)
    resource_path.append(resource_id)
    resource = get_object_at(aws_config, resource_path)
    if 'status_path' in callback_args:
        resource_status = get_object_at(resource, callback_args['status_path'])
    else:
        resource_status = None
    # Get list of VPC security groups for the resource
    # TODO: create EC2 classic SG and see how it differs from a VPC SG ... same for RDS Etc...
    sg_base_path = copy.deepcopy(current_path)
    sg_base_path.pop()
    sg_base_path[1] = 'ec2'
    #sg_base_path.append(callback_args)
    sg_base_path.append('security_groups')
    # Issue 89 & 91 : can instances have no security group?
    try:
      for resource_sg in resource[callback_args['sg_list_attribute_name']]:
        # Get security group
        sg_path = copy.deepcopy(sg_base_path)
        sg_path.append(resource_sg[callback_args['sg_id_attribute_name']])
        print('SG path: %s' % sg_path)
        sg = get_object_at(aws_config, sg_path)
        # Add usage information
        manage_dictionary(sg, 'used_by', {})
        manage_dictionary(sg['used_by'], service, {})
        manage_dictionary(sg['used_by'][service], 'resource_type', {})
        manage_dictionary(sg['used_by'][service]['resource_type'], resource_type, {} if resource_status else [])
        if resource_status:
            manage_dictionary(sg['used_by'][service]['resource_type'][resource_type], resource_status, [])
            if not resource_id in sg['used_by'][service]['resource_type'][resource_type][resource_status]:
                sg['used_by'][service]['resource_type'][resource_type][resource_status].append(resource_id)
        else:
                sg['used_by'][service]['resource_type'][resource_type].append(resource_id)
    except Exception as e:
        region = current_path[3]
        vpc_id = current_path[5]
        if vpc_id == ec2_classic and resource_type == 'elbs':
            pass
        else:
            printError('Failed to parse %s in %s in %s' % (resource_type, vpc_id, region))
            printException(e)
